import os

import pymysql

from sqlink import SqliteDatabase, MetaDatabase


class MysqlDatabase(MetaDatabase):
    def __init__(self):
        super().__init__()
        self.update_db_type("mysql")

    def connect(self):
        connection = pymysql.connect(host='localhost',
                                     user='root',
                                     password='123456')
        self.update_connection(connection)
        # 删除数据库（如果存在）
        self.cursor.execute("DROP DATABASE IF EXISTS test_database")
        self.cursor.execute("CREATE DATABASE IF NOT EXISTS test_database")
        # 选择使用新创建的数据库
        self.cursor.execute("USE test_database")


if os.getenv('DB_IMPL') == 'sqlite':
    MetaDatabaseTest = SqliteDatabase
elif os.getenv('DB_IMPL') == 'mysql':
    MetaDatabaseTest = MysqlDatabase



# -*- coding:utf-8 -*-
from dataclasses import asdict, dataclass

import pytest

from test.test_env import MetaDatabaseTest
from sqlink import Entity, Constraint as cs, Dao, DaoFunc, Sql


@Entity(check_type=True)
@dataclass
class Student:
    name: str
    age: int
    phone: int = cs.not_null
    weight: float = 50.0, cs.not_null
    height: float = cs.not_null, cs.unique
    address: str = 'hit', cs.not_null, cs.unique, cs.datatype("varchar(200)")
    student_id: int = cs.primary_key
    if_aged: bool = False


@Dao(entity=Student, fetch_type=dataclass)
class DaoStudent(DaoFunc):

    @Sql("select * from student;")
    def select_all(self):
        pass

    @Sql("select * from student where student_id=?;", fetch_type=Student)
    def select_student_by_id(self, student_id):
        pass

    @Sql("update student set name=? where student_id=?;")
    def update_name_by_id(self, name, student_id):
        pass

    @Sql("delete from student where student_id=?;")
    def delete_student_by_id(self, student_id):
        pass


class Database(MetaDatabaseTest):
    dao_student = DaoStudent()


@pytest.fixture(scope="session")
def database_class():
    return Database


# ====================================================================================================================
def generate_students(start, end):
    return [
        Student(name=f'李华{i}', age=i, phone=123456789,
                weight=50.0, height=100.0 + i, address=f'hit{i}',
                student_id=i)
        for i in range(start, end)
    ]


def insert_data(db: Database, num=100):
    students = generate_students(0, num)
    db.dao_student.insert(students)
    db.commit()
    return students


# ====================================================================================================================
def test_dao_create_table(init_db: Database):
    """创建表"""
    db = init_db
    db.dao_student.create_table()
    db.commit()


def test_insert(init_db):
    """插入数据"""
    db = init_db
    students = generate_students(0, 100)
    for student in students:
        record_id = db.dao_student.insert(entity=student)
        db.commit()

        result = db.dao_student.select_student_by_id(record_id)
        result = result[0]
        assert asdict(result) == asdict(students[result.student_id])


def test_select_all(init_db):
    """查询全部"""
    db = init_db
    students = insert_data(db)
    results = db.dao_student.select_all()
    for result in results:
        assert asdict(result) == asdict(students[result.student_id])


def test_select_student_by_id(init_db):
    """具体查询"""
    db = init_db
    students = insert_data(db)
    for student in students:
        select_id = student.student_id
        result = db.dao_student.select_student_by_id(student_id=select_id)
        assert asdict(result[0]) == asdict(students[select_id])


def test_update_name(init_db):
    """更新"""
    db = init_db
    students = insert_data(db)
    for student in students:
        select_id = student.student_id
        new_name = '好家伙'
        db.dao_student.update_name_by_id(name=new_name, student_id=select_id)
        db.commit()
        result = db.dao_student.select_student_by_id(select_id)
        assert result[0].name == new_name


def test_insert_many(init_db: Database):
    """批量插入"""
    db = init_db
    start = 101
    end = 200
    students2 = generate_students(start=start, end=end)
    db.dao_student.insert(entity=students2)
    db.commit()

    for i in range(start, end):
        result = db.dao_student.select_student_by_id(i)
        result = result[0]
        assert asdict(result) == asdict(students2[i - start])


if __name__ == '__main__':
    pytest.main(["-vv", "--capture=no", __file__])



# -*- coding:utf-8 -*-
from dataclasses import dataclass

import pytest

from test.test_env import MetaDatabaseTest
from sqlink import Entity, Constraint as cs, Dao, DaoFunc


@Entity
@dataclass
class Student:
    name: str
    age: int
    phone: int = cs.not_null
    weight: float = 50.0, cs.not_null
    height: float = cs.not_null, cs.unique
    address: str = 'hit', cs.not_null, cs.unique, cs.datatype("varchar(200)")
    student_id: int = cs.auto_primary_key
    if_aged: bool = False


@Dao(Student)
class DaoStudent(DaoFunc):
    pass


class Database(MetaDatabaseTest):
    dao_student = DaoStudent()


@pytest.fixture(scope="session")
def database_class():
    return Database


# ====================================================================================================================
def test_create_tables(init_db):
    """建库"""
    db = init_db
    db.create_tables()


def test_create(init_db):
    db = init_db
    db.dao_student.create_table()


if __name__ == '__main__':
    pytest.main(["-vv", "--capture=no", __file__])



# -*- coding:utf-8 -*-
from dataclasses import dataclass

import pytest

from test.test_env import MetaDatabaseTest
from sqlink import Entity, Constraint as cs, Dao, DaoFunc


@Entity
@dataclass
class Student_Comment:
    name: str = cs.comment('学生姓名')
    contact_info: float = cs.comment('联系方式')
    time: str = cs.comment('时间')
    abstract: str = cs.comment('摘要')
    student_id: int = cs.auto_primary_key, cs.comment('学生id')


@Dao(Student_Comment)
class DaoStudent(DaoFunc):
    pass


class Database(MetaDatabaseTest):
    dao_student = DaoStudent()


@pytest.fixture(scope="session")
def database_class():
    return Database


# ====================================================================================================================
def test_create_tables(init_db):
    """建库"""
    db = init_db
    db.create_tables()


def test_create_with_comment(init_db):
    """注释功能"""
    db = init_db
    db.dao_student.create_table()


if __name__ == '__main__':
    pytest.main(["-vv", "--capture=no", __file__])



# -*- coding:utf-8 -*-
import os
import shutil
import sqlite3
from dataclasses import dataclass
from typing import Optional

import pymysql
import pytest
from pytest import fixture

from test.test_env import MetaDatabaseTest
from sqlink import Entity, Constraint as cs, Dao, DaoFunc, Insert
from sqlink.dao import Conflict, Sql, DB_TYPE_SQLITE, DB_TYPE_MYSQL


@Entity(check_type=True)
@dataclass
class Student:
    name: str = cs.unique, cs.datatype("varchar(200)")
    age: int = None
    id: int = cs.auto_primary_key


@Dao(Student, fetch_type=Student)
class DaoStudent(DaoFunc):

    @Insert(conflict=Conflict.error)
    def insert_error(self):
        pass

    @Insert(conflict=Conflict.ignore)
    def insert_ignore(self):
        pass

    @Insert(conflict=Conflict.replace)
    def insert_replace(self):
        pass

    @Sql("select * from __ where name=? limit 1;")
    def get_student_by_name(self, name) -> Optional[Student]:
        pass


class Database(MetaDatabaseTest):
    dao_student = DaoStudent()


# ====================================================================================================================
@fixture(scope="function")
def init_db(request):
    db = Database()
    db_dir = "db_folder"
    # 从request对象中获取当前测试文件的路径
    test_file_name = request.module.__file__
    # 提取文件名（不含扩展名）
    file_name = os.path.splitext(os.path.basename(test_file_name))[0]
    db_path = f"{db_dir}/{file_name}.db"
    if db.db_type == "sqlite":
        if os.path.exists(db_path):
            os.remove(db_path)
        db.connect(db_path)
    elif db.db_type == "mysql":
        db.connect()
    db.create_tables()
    yield db
    db.disconnect()
    if db.db_type == "sqlite":
        if os.path.exists(db_path):
            os.remove(db_path)
            if not os.listdir(db_dir):
                # 文件夹为空，执行删除操作
                shutil.rmtree(db_dir)


students = [Student(name='张三', age=10),
            Student(name='张三', age=11),
            Student(name='李四', age=12)]


# ====================================================================================================================
def test_insert_default(init_db: Database):
    db = init_db
    if db.db_type == DB_TYPE_SQLITE:
        with pytest.raises(sqlite3.IntegrityError) as exc_info:
            for student in students:
                db.dao_student.insert(student)
        assert "UNIQUE constraint failed" in str(exc_info.value)
    elif db.db_type == DB_TYPE_MYSQL:
        with pytest.raises(pymysql.err.IntegrityError) as exc_info:
            for student in students:
                db.dao_student.insert(student)
        assert "Duplicate entry" in str(exc_info.value)
        assert "for key 'student.name'" in str(exc_info.value)
    db.rollback()


def test_insert_error(init_db: Database):
    db = init_db
    if db.db_type == DB_TYPE_SQLITE:
        with pytest.raises(sqlite3.IntegrityError) as exc_info:
            for student in students:
                db.dao_student.insert_error(student)
        assert "UNIQUE constraint failed" in str(exc_info.value)
    elif db.db_type == DB_TYPE_MYSQL:
        with pytest.raises(pymysql.err.IntegrityError) as exc_info:
            for student in students:
                db.dao_student.insert_error(student)
        assert "Duplicate entry" in str(exc_info.value)
        assert "for key 'student.name'" in str(exc_info.value)
    db.rollback()


def test_insert_ignore(init_db: Database):
    db = init_db
    for student in students:
        db.dao_student.insert_ignore(student)
    db.commit()

    select_student = students[0]
    student_conflict = db.dao_student.get_student_by_name(name=select_student.name)
    assert student_conflict.age == select_student.age


def test_insert_replace(init_db: Database):
    db = init_db
    for student in students:
        db.dao_student.insert_replace(student)
    db.commit()

    select_student = students[0]
    student_conflict = db.dao_student.get_student_by_name(name=select_student.name)
    assert student_conflict.age == students[1].age


if __name__ == '__main__':
    pytest.main(["-vv", "--capture=no", __file__])



# -*- coding:utf-8 -*-
import os

import pytest

if __name__ == "__main__":
    os.environ['DB_IMPL'] = 'sqlite'
    print("Running tests with SqliteDatabase...")
    pytest.main([".", "-vv", "--ignore=test_env.py"])
    # 清除环境变量，以防对其他进程产生影响
    del os.environ['DB_IMPL']

    os.environ['DB_IMPL'] = 'mysql'
    print("\nRunning tests with MysqlMetaDatabase...")
    pytest.main([".", "-vv", "--ignore=test_env.py"])
    # 清除环境变量
    del os.environ['DB_IMPL']



# -*- coding:utf-8 -*-
import os
import shutil
from dataclasses import dataclass

import pytest

from test.test_env import MetaDatabaseTest
from sqlink.dao import Dao, Sql, DaoFunc
from sqlink.entity import Entity, Constraint as cs


@Entity(check_type=True)
@dataclass
class Student:
    name: str
    age: int
    id: int = cs.auto_primary_key


@Dao(entity=Student)
class DaoStudent(DaoFunc):

    @Sql("delete from student where id=?;")
    def delete_student_by_id(self, student_id):
        pass


@Entity(check_type=True)
@dataclass
class Score:
    score: float
    student_id: int = cs.primary_key, cs.foreign_key(parent_entity=Student,
                                                     parent_field='id',
                                                     delete_link=cs.cascade)


@Dao(entity=Score)
class DaoScore(DaoFunc):

    @Sql("select * from __ where student_id=?;")
    def get_score_by_student_id(self, student_id):
        pass


class Database(MetaDatabaseTest):
    dao_score = DaoScore()
    dao_student = DaoStudent()


# ====================================================================================================================
@pytest.fixture(scope="module")
def init_db(request):
    db = Database()
    db_dir = "db_folder"
    # 从request对象中获取当前测试文件的路径
    test_file_name = request.module.__file__
    # 提取文件名（不含扩展名）
    file_name = os.path.splitext(os.path.basename(test_file_name))[0]
    db_path = f"{db_dir}/{file_name}.db"
    if db.db_type == "sqlite":
        if os.path.exists(db_path):
            os.remove(db_path)
        db.connect(db_path)
        db.execute("PRAGMA foreign_keys = ON;")
    elif db.db_type == "mysql":
        db.connect()
    db.dao_student.create_table()
    db.dao_score.create_table()
    yield db
    db.disconnect()
    if db.db_type == "sqlite":
        if os.path.exists(db_path):
            os.remove(db_path)
            if not os.listdir(db_dir):
                # 文件夹为空，执行删除操作
                shutil.rmtree(db_dir)


students = [Student(name='张三', age=10),
            Student(name='张三', age=11),
            Student(name='李四', age=12)]


# ====================================================================================================================
def test_insert_data(init_db: Database):
    db = init_db
    db.dao_student.insert(students)


def test_foreign_key_delete_cascade(init_db: Database):
    """外键"""
    db = init_db
    select_id = 1
    score = Score(150.0, student_id=select_id)
    db.dao_score.insert(score)
    db.dao_student.delete_student_by_id(student_id=select_id)
    db.commit()
    result = db.dao_score.get_score_by_student_id(select_id)
    assert len(result) == 0


if __name__ == '__main__':
    pytest.main(["-vv", "--capture=no", __file__])



# -*- coding: utf-8 -*-
import os
import sqlite3
from abc import abstractmethod, ABC

from sqlink.dao import DB_TYPE_SQLITE, DB_TYPE_MYSQL


# ====================================================================================================================
class MetaDatabase(ABC):
    """元数据库类，提供库级的操作。

    由于元编程无法使用代码提示，因此采取继承写法。

    Example:
        @Entity
        @dataclass
        class Student:  # 定义一个数据类
            name: str
            student_id: int = Constraint.auto_primary_key

        @Dao(Student)
        class DaoStudent:  # 定义一个数据访问类

            @Sql("select * from student where student_id=?;")
            def get_student(self, student_id):
                pass

        class Database(MetaDatabase):  # 定义一个数据库类，继承元数据库类
            dao1 = DaoStudent()  # 将各个数据访问类实例化为类中静态变量，集中管理，统一对外。
            dao2 = ...
            dao3 = ...

            def __init__(self):
                super().__init__()
                self.update_db_type("sqlite")  # 声明数据库类型

            def connect(self):
                connection = ...
                self.update_connection(connection)


        db = Database()  # 实例化数据库类，并传入数据库路径
        db.connect()  # 连接数据库
        db.create_tables()  # 创建数据表

    """

    def update_connection(self, connection):
        """设置数据库连接"""
        if not connection:
            raise ValueError(
                "数据库连接异常。"
            )
        self.connection = connection
        self.cursor = self.connection.cursor()
        self._update_cursor()

    def update_db_type(self, db_type: str):
        """设置数据库类型"""
        if db_type not in [DB_TYPE_SQLITE, DB_TYPE_MYSQL]:
            raise ValueError(
                "数据库类型设置错误。"
            )
        self.db_type = db_type
        for dao in self.dao_list:
            dao.update_db_type(self.db_type)

    @abstractmethod
    def connect(self, *args, **kwargs):
        """创建数据库连接connection后，必须再调用update_connection方法"""

    def disconnect(self):
        """断开数据库连接"""
        if self.cursor:
            self.cursor.close()
        if self.connection:
            self.connection.close()

    def create_tables(self):
        """当表不存在时，才会创建数据表。因此可以反复调用该方法，而不会产生错误。"""
        for dao in self.dao_list:
            dao.create_table()  # noqa
        self.commit()

    def get_create_table_sql(self) -> list[str]:
        """返回各个entity的建表sql语句"""
        return [dao.get_create_table_sql() for dao in self.dao_list]

    def commit(self):
        """提交事务"""
        self.connection.commit()

    def rollback(self):
        """回滚事务"""
        self.connection.rollback()

    def execute(self, sql: str, args=None):
        """给外部提供的直接执行sql的接口，避免了再调用内部的connection"""
        if args:
            self.cursor.execute(sql, args)
        else:
            self.cursor.execute(sql)

    def _update_cursor(self):
        for dao in self.dao_list:
            dao.update_cursor(cursor=self.cursor)

    def __new__(cls, *args, **kwargs):
        # 获取子类的所有属性
        subclass_attrs = dir(cls)
        # 初步筛选静态变量（不包括方法和特殊属性）
        static_attrs = [attr for attr in subclass_attrs
                        if not callable(getattr(cls, attr)) and not attr.startswith("__")]
        # 根据是否具有entity属性筛选出最终dao属性，以便在类内部可见
        cls.dao_list = [getattr(cls, attr) for attr in static_attrs
                        if hasattr(getattr(cls, attr), "entity")
                        and hasattr(getattr(cls, attr), "update_cursor")]
        return super().__new__(cls)

    def __init__(self):
        self.db_type = None
        self.connection = None
        self.cursor = None


class SqliteDatabase(MetaDatabase):

    def __init__(self):
        super().__init__()
        self.update_db_type("sqlite")

    def connect(self, db_path: str, *args, **kwargs):
        self.__check_path(db_path)
        connection = sqlite3.connect(db_path, *args, **kwargs)
        self.update_connection(connection)

    @staticmethod
    def __check_path(path):
        """检查并确保目录存在"""
        db_folder = os.path.dirname(path)
        if db_folder != '' and not os.path.exists(db_folder):
            os.makedirs(db_folder, exist_ok=True)
